"use strict";
var __awaiter = (this && this.__awaiter) || function (thisArg, _arguments, P, generator) {
    function adopt(value) { return value instanceof P ? value : new P(function (resolve) { resolve(value); }); }
    return new (P || (P = Promise))(function (resolve, reject) {
        function fulfilled(value) { try { step(generator.next(value)); } catch (e) { reject(e); } }
        function rejected(value) { try { step(generator["throw"](value)); } catch (e) { reject(e); } }
        function step(result) { result.done ? resolve(result.value) : adopt(result.value).then(fulfilled, rejected); }
        step((generator = generator.apply(thisArg, _arguments || [])).next());
    });
};
var __importDefault = (this && this.__importDefault) || function (mod) {
    return (mod && mod.__esModule) ? mod : { "default": mod };
};
Object.defineProperty(exports, "__esModule", { value: true });
exports.logMirrorFundInfo = exports.logValidatorInfo = exports.logNetworkInfo = exports.logBalanceInfo = exports.logAddressInfo = exports.initCtxJsonFromOptions = exports.capFeeAt = exports.getOptions = exports.networkFromOptions = exports.contextFromOptions = exports.optOutOfAirdropForDefi = exports.optOutOfAirdropLedger = exports.optOutOfAirdropPrivateKey = exports.cli = void 0;
const ethersV5_1 = require("ethersV5");
const chalk_1 = __importDefault(require("chalk"));
const context_1 = require("./context");
const utils_1 = require("./utils");
const evmAtomicTx_1 = require("./transaction/evmAtomicTx");
const pvmAtomicTx_1 = require("./transaction/pvmAtomicTx");
const addValidator_1 = require("./transaction/addValidator");
const addDelegator_1 = require("./transaction/addDelegator");
const key_1 = require("./ledger/key");
const sign_1 = require("./ledger/sign");
const transaction_1 = require("./forDefi/transaction");
const withdrawal_1 = require("./forDefi/withdrawal");
const output_1 = require("./output");
const contracts_1 = require("./contracts");
const contracts_2 = require("./constants/contracts");
const screen_1 = require("./constants/screen");
const BASE_DERIVATION_PATH = "m/44'/60'/0'/0/0"; // base derivation path for ledger
const FLR = 1e9; // one FLR in nanoFLR
const MAX_TRANSCTION_FEE = FLR;
// mapping from network to symbol
const networkTokenSymbol = {
    "flare": "FLR",
    "songbird": "SGB",
    "costwo": "C2FLR",
    "coston": "CFLR",
    "localflare": "PHT"
};
function cli(program) {
    return __awaiter(this, void 0, void 0, function* () {
        // global configurations
        program
            .option("--network <network>", "Network name (flare|songbird|costwo|coston|localflare)")
            .option("--ledger", "Use ledger to sign transactions")
            .option("--blind", "Blind signing (used for ledger)", true)
            .option("--derivation-path <derivation-path>", "Ledger address derivation path", BASE_DERIVATION_PATH)
            .option("--ctx-file <file>", "Context file as returned by init-ctx", 'ctx.json')
            .option("--env-path <path>", "Path to the .env file")
            .option("--get-hacked", "Use the .env file with the exposed private key");
        // interactive mode
        program
            .command("interactive").description("Interactive mode")
            .action(() => __awaiter(this, void 0, void 0, function* () {
            // this will never run, here just for --help display
        }));
        // context setup
        program
            .command("init-ctx").description("Initialize context file")
            .option("-p, --public-key <public-key>", "Public key of the account")
            .action((options) => __awaiter(this, void 0, void 0, function* () {
            options = getOptions(program, options);
            initCtxJsonFromOptions(options);
        }));
        // information about the network
        program
            .command("info").description("Relevant information")
            .argument("<addresses|balance|network|validators>", "Type of information")
            .action((type) => __awaiter(this, void 0, void 0, function* () {
            const options = getOptions(program, program.opts());
            const ctx = yield contextFromOptions(options);
            if (type == 'addresses') {
                logAddressInfo(ctx);
            }
            else if (type == 'balance') {
                yield logBalanceInfo(ctx);
            }
            else if (type == 'network') {
                logNetworkInfo(ctx);
            }
            else if (type == 'validators') {
                yield logValidatorInfo(ctx);
            }
            else if (type == 'mirror') {
                yield logMirrorFundInfo(ctx);
            }
            else {
                (0, output_1.logError)(`Unknown information type ${type}`);
            }
        }));
        // transaction construction and sending
        program
            .command("transaction").description("Move funds from one chain to another, stake, and delegate")
            .argument("<importCP|exportCP|importPC|exportPC|delegate|stake>", "Type of a crosschain transaction")
            .option("-i, --transaction-id <transaction-id>", "Id of the transaction to finalize")
            .option("-a, --amount <amount>", "Amount to transfer")
            .option("-f, --fee <fee>", "Transaction fee (in FLR)")
            .option("-n, --node-id <nodeId>", "The id of the node to stake/delegate to")
            .option("-s, --start-time <start-time>", "Start time of the staking/delegating process")
            .option("-e, --end-time <end-time>", "End time of the staking/delegating process")
            .option("--nonce <nonce>", "Nonce of the constructed transaction")
            .option("--delegation-fee <delegation-fee>", "Delegation fee defined by the deployed validator", "10")
            .option("--threshold <threshold>", "Threshold of the constructed transaction", "1")
            .action((type, options) => __awaiter(this, void 0, void 0, function* () {
            options = getOptions(program, options);
            const ctx = yield contextFromOptions(options);
            if (options.getHacked) {
                // for future development: users should get notified before the program gets access to their private keys
                yield cliBuildAndSendTxUsingPrivateKey(type, ctx, options);
            }
            else if (options.ledger) {
                yield cliBuildAndSendTxUsingLedger(type, ctx, options, options.blind, options.derivationPath);
            }
            else {
                yield cliBuildUnsignedTxJson(type, ctx, options.transactionId, options);
            }
        }));
        // signed transaction sending
        program
            .command("send").description("Send signed transaction json to the node")
            .option("-i, --transaction-id <transaction-id>", "Id of the transaction to send to the network")
            .action((options) => __awaiter(this, void 0, void 0, function* () {
            options = getOptions(program, options);
            const ctx = yield contextFromOptions(options);
            yield cliSendSignedTxJson(ctx, options.transactionId);
        }));
        // forDefi signing
        program
            .command("forDefi").description("Sign with ForDefi")
            .argument("<sign|fetch>", "Type of a forDefi transaction")
            .option("-i, --transaction-id <transaction-id>", "Id of the transaction to finalize")
            .option("--withdrawal", "Withdrawing funds from c-chain")
            .action((type, options) => __awaiter(this, void 0, void 0, function* () {
            options = getOptions(program, options);
            if (type == 'sign') {
                if (options.withdrawal) {
                    yield signForDefi(options.transactionId, options.ctxFile, true);
                }
                else {
                    yield signForDefi(options.transactionId, options.ctxFile);
                }
            }
            else if (type == 'fetch') {
                if (options.withdrawal) {
                    yield fetchForDefiTx(options.transactionId, true);
                }
                else {
                    yield fetchForDefiTx(options.transactionId);
                }
            }
        }));
        // withdrawal from c-chain
        program
            .command("withdrawal").description("Withdraw funds from c-chain")
            .option("-i, --transaction-id <transaction-id>", "Id of the transaction to finalize")
            .option("-a, --amount <amount>", "Amount to transfer")
            .option("-t, --to <to>", "Address to send funds to")
            .option("--nonce <nonce>", "Nonce of the constructed transaction")
            .option("--send-signed-tx", "Send signed transaction json to the node")
            .action((options) => __awaiter(this, void 0, void 0, function* () {
            options = getOptions(program, options);
            const ctx = yield contextFromOptions(options);
            if (options.sendSignedTx) {
                yield withdraw_useSignature(ctx, options.transactionId);
            }
            else { // create unsigned transaction
                yield withdraw_getHash(ctx, options.to, options.amount, options.transactionId, options.nonce);
            }
        }));
        // Opt-out from Airdrop
        program
            .command("opt-out").description("Opt-out from the Flare Drop i.e execute optOutOfAirdrop function on DistributionToDelegators smart contract")
            .option("-i, --transaction-id <transaction-id>", "Id of the transaction to finalize")
            .action((options) => __awaiter(this, void 0, void 0, function* () {
            options = getOptions(program, options);
            const ctx = yield contextFromOptions(options);
            if (options.getHacked) {
                yield optOutOfAirdropPrivateKey(Object.keys(screen_1.walletConstants)[2], ctx);
            }
            else if (options.ledger) {
                yield optOutOfAirdropLedger(Object.keys(screen_1.walletConstants)[0], ctx.cAddressHex, options.derivationPath, options.network);
            }
            else { // for ForDefi
                yield optOutOfAirdropForDefi(Object.keys(screen_1.walletConstants)[1], options.transactionId, ctx);
            }
        }));
        // sign and submit smart contract transaction
        program
            .command("signAndSubmit").description("Sign a transaction using private key and submit to chain")
            .option("-i, --transaction-id <transaction-id>", "Id of the transaction to finalize")
            .action((options) => __awaiter(this, void 0, void 0, function* () {
            options = getOptions(program, options);
            const ctx = yield contextFromOptions(options);
            yield signAndSend(ctx, options.network, options.transactionId);
        }));
    });
}
exports.cli = cli;
function optOutOfAirdropPrivateKey(wallet, ctx) {
    return __awaiter(this, void 0, void 0, function* () {
        const optOutParams = {
            cAddress: ctx.cAddressHex,
            network: ctx.config.hrp,
            wallet: wallet,
            pvtKey: ctx.privkHex
        };
        yield (0, contracts_1.optOutOfAirdrop)(optOutParams);
        console.log(chalk_1.default.green("Successfully opted out"));
    });
}
exports.optOutOfAirdropPrivateKey = optOutOfAirdropPrivateKey;
function optOutOfAirdropLedger(wallet, ctxCAddress, ctxDerivationPath, ctxNetwork) {
    return __awaiter(this, void 0, void 0, function* () {
        const optOutParams = {
            cAddress: ctxCAddress,
            network: ctxNetwork,
            wallet: wallet,
            derivationPath: ctxDerivationPath
        };
        console.log("Please sign the transaction on your ledger");
        yield (0, contracts_1.optOutOfAirdrop)(optOutParams);
        console.log(chalk_1.default.green("Successfully opted out"));
    });
}
exports.optOutOfAirdropLedger = optOutOfAirdropLedger;
function optOutOfAirdropForDefi(wallet, transactionId, ctx) {
    return __awaiter(this, void 0, void 0, function* () {
        const optOutParams = {
            cAddress: ctx.cAddressHex,
            network: ctx.config.hrp,
            wallet: wallet,
            transactionId: transactionId
        };
        yield (0, contracts_1.optOutOfAirdrop)(optOutParams);
    });
}
exports.optOutOfAirdropForDefi = optOutOfAirdropForDefi;
/**
 * @description - returns context from the options that are passed
 * @param options - option to define whether its from ledger/env/ctx.file
 * @returns Returns the context based the source passed in the options
 */
function contextFromOptions(options) {
    return __awaiter(this, void 0, void 0, function* () {
        if (options.ledger) {
            (0, output_1.logInfo)("Fetching account from ledger...");
            const account = yield (0, key_1.ledgerGetAccount)(options.derivationPath, options.network);
            const ctx = (0, context_1.getContext)(options.network, account.publicKey);
            return ctx;
        }
        else if (options.envPath) {
            return (0, context_1.contextEnv)(options.envPath, options.network);
        }
        else {
            return (0, context_1.contextFile)(options.ctxFile);
        }
    });
}
exports.contextFromOptions = contextFromOptions;
// Network is obtained from context file, if it exists, else from --network flag.
// This is because ledger does not need a context file
/**
 * @description Returns the network config from the options that were passed
 * @param options - contains the options to derive the network
 * @returns - network
 */
function networkFromOptions(options) {
    let network = options.network;
    if (network == undefined) {
        try {
            network = (0, context_1.networkFromContextFile)(options.ctxFile);
        }
        catch (e) {
            network = "flare";
        }
    }
    (0, output_1.logInfo)(`Using network: ${network}`);
    return network;
}
exports.networkFromOptions = networkFromOptions;
/**
 * @description - Returns the options for a command
 * @param program - the command
 * @param options - option available for the command
 */
function getOptions(program, options) {
    const allOptions = Object.assign(Object.assign({}, program.opts()), options);
    const network = networkFromOptions(allOptions);
    // amount and fee are given in FLR, transform into nanoFLR (FLR = 1e9 nanoFLR)
    if (allOptions.amount) {
        allOptions.amount = (0, utils_1.decimalToInteger)(allOptions.amount.replace(/,/g, ''), 9);
    }
    if (allOptions.fee) {
        allOptions.fee = (0, utils_1.decimalToInteger)(allOptions.fee, 9);
    }
    return Object.assign(Object.assign({}, allOptions), { network });
}
exports.getOptions = getOptions;
/**
 * @description - Returms the fee getting used
 * @param cap - the max allowed free
 * @param usedFee - fee that was used
 * @param specifiedFee - fee specified by the user
 */
function capFeeAt(cap, network, usedFee, specifiedFee) {
    if (usedFee !== undefined && usedFee !== specifiedFee) { // if usedFee was specified by the user, we don't cap it
        const usedFeeNumber = Number(usedFee); // if one of the fees is defined, usedFee is defined
        const symbol = networkTokenSymbol[network];
        if (usedFeeNumber > cap)
            throw new Error(`Used fee of ${usedFeeNumber / FLR} ${symbol} is higher than the maximum allowed fee of ${cap / FLR} ${symbol}`);
        (0, output_1.logInfo)(`Using fee of ${usedFeeNumber / FLR} ${symbol}`);
    }
}
exports.capFeeAt = capFeeAt;
//////////////////////////////////////////////////////////////////////////////////////////
// transaction-type translators
function buildUnsignedTxJson(transactionType, ctx, params) {
    switch (transactionType) {
        case 'exportCP': {
            return (0, evmAtomicTx_1.getUnsignedExportTxCP)(ctx, (0, utils_1.toBN)(params.amount), (0, utils_1.toBN)(params.fee), (params.nonce === undefined) ? undefined : Number(params.nonce));
        }
        case 'importCP':
            return (0, pvmAtomicTx_1.getUnsignedImportTxCP)(ctx, Number(params.threshold));
        case 'exportPC': {
            return (0, pvmAtomicTx_1.getUnsignedExportTxPC)(ctx, (0, utils_1.toBN)(params.amount), Number(params.threshold));
        }
        case 'importPC': {
            return (0, evmAtomicTx_1.getUnsignedImportTxPC)(ctx, (0, utils_1.toBN)(params.fee));
        }
        case 'stake': {
            return (0, addValidator_1.getUnsignedAddValidator)(ctx, params.nodeId, (0, utils_1.toBN)(params.amount), (0, utils_1.toBN)(params.startTime), (0, utils_1.toBN)(params.endTime), Number(params.delegationFee), Number(params.threshold));
        }
        case 'delegate': {
            return (0, addDelegator_1.getUnsignedAddDelegator)(ctx, params.nodeId, (0, utils_1.toBN)(params.amount), (0, utils_1.toBN)(params.startTime), (0, utils_1.toBN)(params.endTime), Number(params.threshold));
        }
        default:
            throw new Error(`Unknown transaction type: ${transactionType}`);
    }
}
function sendSignedTxJson(ctx, signedTxJson) {
    return __awaiter(this, void 0, void 0, function* () {
        switch (signedTxJson.transactionType) {
            case 'exportCP': {
                const { chainTxId } = yield (0, evmAtomicTx_1.issueSignedEvmTxCPExport)(ctx, signedTxJson);
                return chainTxId;
            }
            case 'importPC': {
                const { chainTxId } = yield (0, evmAtomicTx_1.issueSignedEvmTxPCImport)(ctx, signedTxJson);
                return chainTxId;
            }
            case 'exportPC':
            case 'importCP':
            case 'stake':
            case 'delegate': {
                const { chainTxId } = yield (0, pvmAtomicTx_1.issueSignedPvmTx)(ctx, signedTxJson);
                return chainTxId;
            }
            default:
                throw new Error(`Unknown transaction type: ${signedTxJson.transactionType}`);
        }
    });
}
function buildAndSendTxUsingPrivateKey(transactionType, ctx, params) {
    return __awaiter(this, void 0, void 0, function* () {
        if (transactionType === 'exportCP') {
            return (0, evmAtomicTx_1.exportTxCP)(ctx, (0, utils_1.toBN)(params.amount), (0, utils_1.toBN)(params.fee));
        }
        else if (transactionType === 'importCP') {
            return (0, pvmAtomicTx_1.importTxCP)(ctx, Number(params.threshold));
        }
        else if (transactionType === 'exportPC') {
            return (0, pvmAtomicTx_1.exportTxPC)(ctx, (0, utils_1.toBN)(params.amount), Number(params.threshold));
        }
        else if (transactionType === 'importPC') {
            return (0, evmAtomicTx_1.importTxPC)(ctx, (0, utils_1.toBN)(params.fee));
        }
        else if (transactionType === 'stake') {
            return (0, addValidator_1.addValidator)(ctx, params.nodeId, (0, utils_1.toBN)(params.amount), (0, utils_1.toBN)(params.startTime), (0, utils_1.toBN)(params.endTime), Number(params.delegationFee), Number(params.threshold));
        }
        else if (transactionType === 'delegate') {
            return (0, addDelegator_1.addDelegator)(ctx, params.nodeId, (0, utils_1.toBN)(params.amount), (0, utils_1.toBN)(params.startTime), (0, utils_1.toBN)(params.endTime), Number(params.threshold));
        }
        else {
            throw new Error(`Unknown transaction type ${transactionType}`);
        }
    });
}
//////////////////////////////////////////////////////////////////////////////////////////
// initializing ctx.json
function initCtxJsonFromOptions(options, derivationPath = BASE_DERIVATION_PATH) {
    return __awaiter(this, void 0, void 0, function* () {
        let ctxFile;
        if (options.ledger) {
            const { publicKey, address } = yield (0, key_1.ledgerGetAccount)(derivationPath, options.network);
            const ethAddress = (0, utils_1.publicKeyToEthereumAddressString)(publicKey);
            ctxFile = { wallet: "ledger", publicKey, ethAddress, flareAddress: address, network: options.network, derivationPath };
        }
        else if (options.publicKey) {
            if (!(0, utils_1.validatePublicKey)(options.publicKey))
                return (0, output_1.logError)('Invalid public key');
            ctxFile = { wallet: "publicKey", publicKey: options.publicKey, network: options.network };
            if (options.vaultId) {
                ctxFile = Object.assign(Object.assign({}, ctxFile), { vaultId: options.vaultId });
            }
        }
        else {
            throw new Error('Either --ledger or --public-key must be specified');
        }
        (0, utils_1.initCtxJson)(ctxFile);
        (0, output_1.logSuccess)("Context file created");
    });
}
exports.initCtxJsonFromOptions = initCtxJsonFromOptions;
//////////////////////////////////////////////////////////////////////////////////////////
// Network info
/**
 * @description Logs the address info
 * @param ctx - the context file aka ctx.json
 * @returns Returns the address info
 */
function logAddressInfo(ctx) {
    const [pubX, pubY] = ctx.publicKey;
    const compressedPubKey = (0, utils_1.compressPublicKey)(pubX, pubY).toString('hex');
    (0, output_1.logInfo)(`Addresses on the network "${ctx.config.hrp}"`);
    (0, output_1.log)(`P-chain address: ${ctx.pAddressBech32}`);
    (0, output_1.log)(`C-chain address hex: ${ctx.cAddressHex}`);
    (0, output_1.log)(`secp256k1 public key: 0x${compressedPubKey}`);
}
exports.logAddressInfo = logAddressInfo;
/**
 * @description Logs the balance info of the account
 * @param ctx - the context file aka ctx.json
 */
function logBalanceInfo(ctx) {
    return __awaiter(this, void 0, void 0, function* () {
        let cbalance = ((0, utils_1.toBN)(yield ctx.web3.eth.getBalance(ctx.cAddressHex))).toString();
        let pbalance = ((0, utils_1.toBN)((yield ctx.pchain.getBalance(ctx.pAddressBech32)).balance)).toString();
        cbalance = (0, utils_1.integerToDecimal)(cbalance, 18);
        pbalance = (0, utils_1.integerToDecimal)(pbalance, 9);
        const symbol = networkTokenSymbol[ctx.config.hrp];
        (0, output_1.logInfo)(`Balances on the network "${ctx.config.hrp}"`);
        (0, output_1.log)(`C-chain ${ctx.cAddressHex}: ${cbalance} ${symbol}`);
        (0, output_1.log)(`P-chain ${ctx.pAddressBech32}: ${pbalance} ${symbol}`);
    });
}
exports.logBalanceInfo = logBalanceInfo;
/**
 * @description Logs info aboout P,C and asset id
 * @param ctx - the context file
 */
function logNetworkInfo(ctx) {
    const pchainId = ctx.pchain.getBlockchainID();
    const cchainId = ctx.cchain.getBlockchainID();
    (0, output_1.logInfo)(`Information about the network "${ctx.config.hrp}"`);
    (0, output_1.log)(`blockchainId for P-chain: ${pchainId}`);
    (0, output_1.log)(`blockchainId for C-chain: ${cchainId}`);
    (0, output_1.log)(`assetId: ${ctx.avaxAssetID}`);
}
exports.logNetworkInfo = logNetworkInfo;
/**
 * @description Logs the validator information regrading current and pending validators
 * @param ctx - the context file
 */
function logValidatorInfo(ctx) {
    return __awaiter(this, void 0, void 0, function* () {
        const pending = yield ctx.pchain.getPendingValidators();
        const current = yield ctx.pchain.getCurrentValidators();
        const fpending = JSON.stringify(pending, null, 2);
        const fcurrent = JSON.stringify(current, null, 2);
        (0, output_1.logInfo)(`Validators on the network "${ctx.config.hrp}"`);
        (0, output_1.log)(`pending: ${fpending}`);
        (0, output_1.log)(`current: ${fcurrent}`);
    });
}
exports.logValidatorInfo = logValidatorInfo;
/**
 * @description Logs mirror fund details
 * @param ctx - context
 */
function logMirrorFundInfo(ctx) {
    return __awaiter(this, void 0, void 0, function* () {
        const mirroFundDetails = yield (0, contracts_1.fetchMirrorFunds)(ctx);
        (0, output_1.logInfo)(`Mirror fund details on the network "${ctx.config.hrp}"`);
        (0, output_1.log)(JSON.stringify(mirroFundDetails, null, 2));
    });
}
exports.logMirrorFundInfo = logMirrorFundInfo;
//////////////////////////////////////////////////////////////////////////////////////////
// Transaction building and execution
function cliBuildAndSendTxUsingLedger(transactionType, ctx, params, blind, derivationPath) {
    return __awaiter(this, void 0, void 0, function* () {
        if (transactionType === "exportCP" || transactionType === "exportPC") {
            (0, output_1.logInfo)("Creating export transaction...");
        }
        if (transactionType === "importCP" || transactionType === "importPC") {
            (0, output_1.logInfo)("Creating import transaction...");
        }
        const unsignedTxJson = yield buildUnsignedTxJson(transactionType, ctx, params);
        capFeeAt(MAX_TRANSCTION_FEE, ctx.config.hrp, unsignedTxJson.usedFee, params.fee);
        if (blind) {
            const filename = unsignedTxJson.signatureRequests[0].message.slice(0, 6);
            (0, utils_1.saveUnsignedTxJson)(unsignedTxJson, filename, 'proofs');
            (0, output_1.logWarning)(`Blind signing! Validate generated proofs/${filename}.unsignedTx.json file.`);
        }
        (0, output_1.logInfo)(`Please review and sign transaction on your ledger device...`);
        const { signature } = yield (0, sign_1.ledgerSign)(unsignedTxJson, derivationPath, blind);
        const signedTxJson = Object.assign(Object.assign({}, unsignedTxJson), { signature });
        (0, output_1.logInfo)("Sending transaction to the node...");
        const chainTxId = yield sendSignedTxJson(ctx, signedTxJson);
        (0, output_1.logSuccess)(`Transaction with hash ${chainTxId} sent to the node`);
    });
}
function cliBuildUnsignedTxJson(transactionType, ctx, id, params) {
    return __awaiter(this, void 0, void 0, function* () {
        const unsignedTxJson = yield buildUnsignedTxJson(transactionType, ctx, params);
        capFeeAt(MAX_TRANSCTION_FEE, ctx.config.hrp, unsignedTxJson.usedFee, params.fee);
        (0, utils_1.saveUnsignedTxJson)(unsignedTxJson, id);
        (0, output_1.logSuccess)(`Unsigned transaction ${id} constructed`);
    });
}
function cliSendSignedTxJson(ctx, id) {
    return __awaiter(this, void 0, void 0, function* () {
        if ((0, utils_1.isAlreadySentToChain)(id)) {
            throw new Error("Tx already sent to chain");
        }
        const signedTxnJson = (0, utils_1.readSignedTxJson)(id);
        let chainTxId;
        if (signedTxnJson.transactionType === contracts_2.contractTransactionName) {
            chainTxId = yield (0, contracts_1.submitForDefiTxn)(id, signedTxnJson.signature, ctx.config.hrp);
        }
        else {
            chainTxId = yield sendSignedTxJson(ctx, signedTxnJson);
        }
        (0, utils_1.addFlagForSentSignedTx)(id);
        (0, output_1.logSuccess)(`Signed transaction ${id} with hash ${chainTxId} sent to the node`);
    });
}
function cliBuildAndSendTxUsingPrivateKey(transactionType, ctx, params) {
    return __awaiter(this, void 0, void 0, function* () {
        const { txid, usedFee } = yield buildAndSendTxUsingPrivateKey(transactionType, ctx, params);
        const symbol = networkTokenSymbol[ctx.config.hrp];
        if (usedFee)
            (0, output_1.logInfo)(`Used fee of ${(0, utils_1.integerToDecimal)(usedFee, 9)} ${symbol}`);
        (0, output_1.logSuccess)(`Transaction with hash ${txid} built and sent to the network`);
    });
}
//////////////////////////////////////////////////////////////////////////////////////////
// Transaction execution using ForDefi api
function signForDefi(transaction, ctx, withdrawal = false) {
    return __awaiter(this, void 0, void 0, function* () {
        const txid = yield (0, transaction_1.sendToForDefi)(transaction, ctx, withdrawal);
        (0, output_1.logSuccess)(`Transaction with hash ${txid} sent to the ForDefi`);
    });
}
function fetchForDefiTx(transaction, withdrawal = false) {
    return __awaiter(this, void 0, void 0, function* () {
        if ((0, utils_1.isAlreadySentToChain)(transaction)) {
            throw new Error("Tx already sent to chain");
        }
        const signature = yield (0, transaction_1.getSignature)(transaction, withdrawal);
        (0, output_1.logSuccess)(`Success! Signature: ${signature}`);
    });
}
function withdraw_getHash(ctx, to, amount, id, nonce) {
    return __awaiter(this, void 0, void 0, function* () {
        const fileId = yield (0, withdrawal_1.createWithdrawalTransaction)(ctx, to, amount, id, nonce);
        (0, output_1.logSuccess)(`Transaction ${fileId} constructed`);
    });
}
function withdraw_useSignature(ctx, id) {
    return __awaiter(this, void 0, void 0, function* () {
        const txId = yield (0, withdrawal_1.sendSignedWithdrawalTransaction)(ctx, id);
        (0, output_1.logSuccess)(`Transaction ${txId} sent to the node`);
    });
}
//////////////////////////////////////////////////////////////////////////////////////////
// smart contract transaction signing
function signAndSend(ctx, network, id) {
    return __awaiter(this, void 0, void 0, function* () {
        const tx = (0, utils_1.readUnsignedTxJson)(id);
        if (!tx)
            throw new Error("Invalid txn file");
        const valueStr = tx.rawTx.value;
        const valueBN = ethersV5_1.BigNumber.from(valueStr);
        tx.rawTx.value = valueBN;
        const provider = new ethersV5_1.ethers.providers.JsonRpcProvider((0, context_1.rpcUrlFromNetworkConfig)(network));
        const wallet = new ethersV5_1.ethers.Wallet(ctx.privkHex);
        const signedTx = yield wallet.signTransaction(tx.rawTx);
        const chainId = yield provider.sendTransaction(signedTx);
        (0, output_1.logSuccess)(`Signed transaction ${id} with hash ${chainId.hash} sent to the node`);
    });
}
