"use strict";
var __importDefault = (this && this.__importDefault) || function (mod) {
    return (mod && mod.__esModule) ? mod : { "default": mod };
};
Object.defineProperty(exports, "__esModule", { value: true });
exports.context = exports.getNetworkConfig = exports.getContext = exports.networkFromContextFile = exports.contextFile = exports.contextEnv = exports.readContextFile = exports.rpcUrlFromNetworkConfig = void 0;
const fs_1 = __importDefault(require("fs"));
const web3_1 = __importDefault(require("web3"));
const flarejs_1 = require("@flarenetwork/flarejs");
const utils_1 = require("@flarenetwork/flarejs/dist/utils");
const network_1 = require("./constants/network");
const utils_2 = require("./utils");
/**
 * @param network
 * @returns {string} The RPC url of the given network
 */
function rpcUrlFromNetworkConfig(network) {
    const config = getNetworkConfig(network);
    return `${config.protocol}://${config.ip}/ext/bc/C/rpc`;
}
exports.rpcUrlFromNetworkConfig = rpcUrlFromNetworkConfig;
/**
 * @description - parses the file and returns the context of ctx.json
 * @param ctxFile - path to the context file
 * @returns - context
 */
function readContextFile(ctxFile) {
    const file = fs_1.default.readFileSync(ctxFile, 'utf8');
    return JSON.parse(file);
}
exports.readContextFile = readContextFile;
/**
 * @description Returns the context from .env file
 * @param path - path to the .env file
 * @param network - network info. can be localflare, costwo, flare
 * @returns - returns the context from .env file
 */
function contextEnv(path, network) {
    require('dotenv').config({ path: path });
    return getContext(network, process.env.PUBLIC_KEY, process.env.PRIVATE_KEY_HEX, process.env.PRIVATE_KEY_CB58);
}
exports.contextEnv = contextEnv;
/**
 * @description - returns context from the file
 * @param ctxFile - path to the context file
 * @returns returns the context
 */
function contextFile(ctxFile) {
    const ctx = readContextFile(ctxFile);
    return getContext(ctx.network, ctx.publicKey);
}
exports.contextFile = contextFile;
/**
 * @description - returns the network from the context file
 * @param ctxFile - context file
 * @returns returns the network from the context
 */
function networkFromContextFile(ctxFile) {
    const ctx = readContextFile(ctxFile);
    return ctx.network;
}
exports.networkFromContextFile = networkFromContextFile;
/**
 * @description returns the context
 * @param network - network name: flare/localflare/costwo
 * @param publicKey - public key
 * @param privateKeyHex - private key in hex format
 * @param privateKeyCB58 - private key in cb58 format
 * @returns context
 */
function getContext(network, publicKey, privateKeyHex, privateKeyCB58) {
    return context(getNetworkConfig(network), publicKey, privateKeyHex, privateKeyCB58);
}
exports.getContext = getContext;
/**
 * @description - returns the network config based on network that was passed
 * @param network - network name: flare/localflare/costwo
 * @returns the network configuration
 */
function getNetworkConfig(network) {
    let networkConfig;
    if (network == 'flare' || network === undefined) {
        networkConfig = network_1.flare;
    }
    else if (network == 'songbird') {
        networkConfig = network_1.songbird;
    }
    else if (network == 'costwo') {
        networkConfig = network_1.costwo;
    }
    else if (network == 'coston') {
        networkConfig = network_1.coston;
    }
    else if (network == 'localflare') {
        networkConfig = network_1.localflare;
    }
    else if (network == 'local') {
        networkConfig = network_1.local;
    }
    else
        throw Error('Invalid network');
    return networkConfig;
}
exports.getNetworkConfig = getNetworkConfig;
/**
 * The main function that returns the cotext
 * @param config - network configuration
 * @param publicKey - public key
 * @param privkHex - private key in hex format
 * @param privkCB58 - private key in cb58 format
 * @returns the context object
 */
function context(config, publicKey, privkHex, privkCB58) {
    const { protocol, ip, port, networkID } = config;
    // those two addresses should be derived for most cli applications
    let cAddressHex;
    let addressBech32;
    // derive private key in both cb58 and hex if only one is provided
    const bintools = flarejs_1.BinTools.getInstance();
    if (privkHex !== undefined && privkHex !== '') {
        privkHex = (0, utils_2.unPrefix0x)(privkHex);
        const privkBuf = bintools.addChecksum(flarejs_1.Buffer.from(privkHex, 'hex'));
        privkCB58 = bintools.bufferToB58(privkBuf);
    }
    else if (privkCB58 !== undefined && privkCB58 !== '') {
        const privkBuf = bintools.cb58Decode(privkCB58);
        privkHex = privkBuf.toString('hex');
    }
    // derive the public key coords if private key is present and check that they match
    // the public key if provided
    let publicKeyPair;
    if (publicKey) {
        publicKeyPair = (0, utils_2.decodePublicKey)(publicKey);
        publicKey = "04" + Buffer.concat(publicKeyPair).toString('hex'); // standardize
    }
    if (privkHex) {
        const [pubX, pubY] = (0, utils_2.privateKeyToPublicKey)(Buffer.from(privkHex, 'hex'));
        if (publicKey && (!publicKeyPair[0].equals(pubX) || !publicKeyPair[1].equals(pubY))) {
            throw Error("provided private key does not match the public key");
        }
        publicKeyPair = [pubX, pubY];
    }
    const path = '/ext/bc/C/rpc';
    const iport = port ? `${ip}:${port}` : `${ip}`;
    const rpcurl = `${protocol}://${iport}`;
    const web3 = new web3_1.default(`${rpcurl}${path}`);
    // derive addresses from public key if provided (bech32 is later derived again)
    if (publicKey) {
        cAddressHex = (0, utils_2.publicKeyToEthereumAddressString)(publicKey);
        cAddressHex = web3.utils.toChecksumAddress(cAddressHex); // add checksum
        addressBech32 = (0, utils_2.publicKeyToBech32AddressString)(publicKey, config.hrp);
    }
    const avalanche = new flarejs_1.Avalanche(ip, port, protocol, networkID);
    const cchain = avalanche.CChain();
    const pchain = avalanche.PChain();
    const cKeychain = cchain.keyChain();
    const pKeychain = pchain.keyChain();
    if (privkCB58 || publicKey) {
        const key = (privkCB58) ? `${utils_1.PrivateKeyPrefix}${privkCB58}` : `${utils_1.PublicKeyPrefix}${publicKey}`;
        pKeychain.importKey(key);
        cKeychain.importKey(key);
    }
    const pAddressStrings = pchain.keyChain().getAddressStrings();
    const cAddressStrings = cchain.keyChain().getAddressStrings();
    const pAddressBech32 = pAddressStrings[0] || `P-${addressBech32}`;
    const cAddressBech32 = cAddressStrings[0] || `C-${addressBech32}`;
    if (privkHex) {
        const cAccount = web3.eth.accounts.privateKeyToAccount(privkHex);
        const _cAddressHex = cAccount.address;
        if (cAddressHex && cAddressHex.toLowerCase() !== _cAddressHex.toLowerCase()) {
            throw Error('c-chain address does not match private key');
        }
        cAddressHex = _cAddressHex;
    }
    const pChainBlockchainID = utils_1.Defaults.network[networkID].P.blockchainID;
    const cChainBlockchainID = utils_1.Defaults.network[networkID].C.blockchainID;
    const avaxAssetID = utils_1.Defaults.network[networkID].P.avaxAssetID;
    return {
        privkHex: privkHex,
        privkCB58: privkCB58,
        publicKey: publicKeyPair,
        rpcurl: rpcurl,
        web3: web3,
        avalanche: avalanche,
        cchain: cchain,
        pchain: pchain,
        cKeychain: cKeychain,
        pKeychain: pKeychain,
        pAddressBech32: pAddressBech32,
        cAddressBech32: cAddressBech32,
        cAddressHex: cAddressHex,
        cChainBlockchainID: cChainBlockchainID,
        pChainBlockchainID: pChainBlockchainID,
        avaxAssetID: avaxAssetID,
        config: config
    };
}
exports.context = context;
