"use strict";
var __awaiter = (this && this.__awaiter) || function (thisArg, _arguments, P, generator) {
    function adopt(value) { return value instanceof P ? value : new P(function (resolve) { resolve(value); }); }
    return new (P || (P = Promise))(function (resolve, reject) {
        function fulfilled(value) { try { step(generator.next(value)); } catch (e) { reject(e); } }
        function rejected(value) { try { step(generator["throw"](value)); } catch (e) { reject(e); } }
        function step(result) { result.done ? resolve(result.value) : adopt(result.value).then(fulfilled, rejected); }
        step((generator = generator.apply(thisArg, _arguments || [])).next());
    });
};
var __importDefault = (this && this.__importDefault) || function (mod) {
    return (mod && mod.__esModule) ? mod : { "default": mod };
};
Object.defineProperty(exports, "__esModule", { value: true });
exports.optOutOfAirdrop = exports.fetchMirrorFunds = exports.submitForDefiTxn = exports.claimRewards = exports.registerAddress = exports.isUnclaimedReward = exports.isAddressRegistered = void 0;
const fs_1 = __importDefault(require("fs"));
const chalk_1 = __importDefault(require("chalk"));
const process_1 = require("process");
const bech32_1 = require("bech32");
const ethersV5_1 = require("ethersV5");
const forDefi_1 = require("./constants/forDefi");
const contracts_1 = require("./constants/contracts");
const screen_1 = require("./constants/screen");
const utils_1 = require("./utils");
const context_1 = require("./context");
const sign_1 = require("./ledger/sign");
/**
 * @description checks if the address is registered with the addressBinder contract
 * @param ethAddressToCheck
 * @param network
 */
function isAddressRegistered(ethAddressToCheck, network) {
    return __awaiter(this, void 0, void 0, function* () {
        console.log("Checking Address Registration...");
        const rpcUrl = (0, context_1.rpcUrlFromNetworkConfig)(network);
        const addressBinderContractAddress = yield getContractAddress(network, contracts_1.addressBinderContractName);
        const provider = new ethersV5_1.ethers.providers.JsonRpcProvider(rpcUrl);
        const abi = (0, contracts_1.getAddressBinderABI)();
        const contract = new ethersV5_1.ethers.Contract(addressBinderContractAddress, abi, provider);
        const result = yield contract.cAddressToPAddress(ethAddressToCheck);
        const pChainAddress = bech32_1.bech32.encode(network, bech32_1.bech32.toWords(Buffer.from(result.slice(2), 'hex')));
        if (result !== '0x0000000000000000000000000000000000000000') {
            console.log(chalk_1.default.green(`Address associated with key ${ethAddressToCheck}: ${pChainAddress}`));
            return true;
        }
        else {
            console.log(chalk_1.default.red(`No address found for key ${ethAddressToCheck}`));
            return false;
        }
    });
}
exports.isAddressRegistered = isAddressRegistered;
/**
 * @description checks if there are any unclaimed rewards with the validatorRewardManager contract
 * @param ethAddressToCheck
 * @param network
 */
function isUnclaimedReward(ethAddressToCheck, network) {
    return __awaiter(this, void 0, void 0, function* () {
        console.log("Checking your Rewards status...");
        const rpcUrl = (0, context_1.rpcUrlFromNetworkConfig)(network);
        const validatorRewardManagerContractAddress = yield getContractAddress(network, contracts_1.validatorRewardManagerContractName);
        const provider = new ethersV5_1.ethers.providers.JsonRpcProvider(rpcUrl);
        const abi = (0, contracts_1.getValidatorRewardManagerABI)();
        const contract = new ethersV5_1.ethers.Contract(validatorRewardManagerContractAddress, abi, provider);
        const rewards = yield contract.getStateOfRewards(ethAddressToCheck);
        const totalRewardNumber = rewards[0];
        const claimedRewardNumber = rewards[1];
        const unclaimedRewards = totalRewardNumber.sub(claimedRewardNumber);
        if (unclaimedRewards.gt(ethersV5_1.BigNumber.from("0"))) {
            const unclaimedRewardsInFLR = unclaimedRewards.div(ethersV5_1.ethers.constants.WeiPerEther); // div by 1e18
            console.log(chalk_1.default.green(`You have unclaimed rewards worth ${unclaimedRewardsInFLR} FLR`));
            return true;
        }
        else {
            console.log(chalk_1.default.red(`No unclaimed rewards found`));
            return false;
        }
    });
}
exports.isUnclaimedReward = isUnclaimedReward;
/**
 * @description cregisters address  with the addressBinder contract
 * @param {RegisterAddressInterface} addressParams
 */
function registerAddress(addressParams) {
    return __awaiter(this, void 0, void 0, function* () {
        const { publicKey, pAddress, cAddress, network, wallet, derivationPath, pvtKey, transactionId } = addressParams;
        const rpcUrl = (0, context_1.rpcUrlFromNetworkConfig)(network);
        const addressBinderContractAddress = yield getContractAddress(network, contracts_1.addressBinderContractName);
        const provider = new ethersV5_1.ethers.providers.JsonRpcProvider(rpcUrl);
        const abi = (0, contracts_1.getAddressBinderABI)();
        const contract = new ethersV5_1.ethers.Contract(addressBinderContractAddress, abi, provider);
        const checksumAddress = ethersV5_1.ethers.utils.getAddress(cAddress);
        const nonce = yield provider.getTransactionCount(checksumAddress);
        const config = (0, context_1.getNetworkConfig)(network);
        const pAddr = (0, utils_1.prefix0x)(Buffer.from(bech32_1.bech32.fromWords(bech32_1.bech32.decode(pAddress.slice(2)).words)).toString('hex'));
        const publicKeyPrefixed = (0, utils_1.prefix0x)(publicKey);
        const gasEstimate = yield contract.estimateGas.registerAddresses(publicKeyPrefixed, pAddr, cAddress, { from: cAddress });
        const gasPrice = yield provider.getGasPrice();
        const populatedTx = yield contract.populateTransaction.registerAddresses(publicKeyPrefixed, pAddr, cAddress);
        const unsignedTx = Object.assign(Object.assign({}, populatedTx), { nonce, chainId: config.networkID, gasPrice, gasLimit: gasEstimate });
        yield signContractTransaction(wallet, unsignedTx, contract, derivationPath, transactionId, pvtKey);
    });
}
exports.registerAddress = registerAddress;
/**
 * @description claims rewards from the ValidatorRewardManager Contract
 * @param {ClaimRewardsInterface} rewardsParams
 */
function claimRewards(rewardsParams) {
    return __awaiter(this, void 0, void 0, function* () {
        const { claimAmount, ownerAddress, receiverAddress, network, wallet, derivationPath, pvtKey, transactionId } = rewardsParams;
        const rewardAmount = ethersV5_1.BigNumber.from(claimAmount).mul(ethersV5_1.ethers.constants.WeiPerEther);
        if (rewardAmount.gt(yield getUnclaimedRewards(ownerAddress, network)) || !rewardAmount.gt(ethersV5_1.BigNumber.from("0"))) {
            throw new Error("Incorrect amount to claim");
        }
        const rpcUrl = (0, context_1.rpcUrlFromNetworkConfig)(network);
        const validatorRewardManagerContractAddress = yield getContractAddress(network, contracts_1.validatorRewardManagerContractName);
        const provider = new ethersV5_1.ethers.providers.JsonRpcProvider(rpcUrl);
        const abi = (0, contracts_1.getValidatorRewardManagerABI)();
        const contract = new ethersV5_1.ethers.Contract(validatorRewardManagerContractAddress, abi, provider);
        const checksumAddress = ethersV5_1.ethers.utils.getAddress(ownerAddress);
        const nonce = yield provider.getTransactionCount(checksumAddress);
        const config = (0, context_1.getNetworkConfig)(network);
        let gasEstimate;
        try {
            gasEstimate = yield contract.estimateGas.claim(ownerAddress, (0, utils_1.prefix0x)(receiverAddress), rewardAmount, false, { from: ownerAddress });
        }
        catch (_a) {
            console.log(chalk_1.default.red("Incorrect arguments passed"));
            (0, process_1.exit)();
        }
        const gasPrice = yield provider.getGasPrice();
        const populatedTx = yield contract.populateTransaction.claim(ownerAddress, (0, utils_1.prefix0x)(receiverAddress), rewardAmount, false);
        const unsignedTx = Object.assign(Object.assign({}, populatedTx), { nonce, chainId: config.networkID, gasPrice, gasLimit: gasEstimate });
        yield signContractTransaction(wallet, unsignedTx, contract, derivationPath, transactionId, pvtKey);
    });
}
exports.claimRewards = claimRewards;
/**
 * @description submits signed ForDefi txn to the chain
 * @param id Id of the txn to submit
 * @param signature signature of the signed txn from ForDefi
 * @param network
 * @returns TxnHash
 */
function submitForDefiTxn(id, signature, network) {
    return __awaiter(this, void 0, void 0, function* () {
        const provider = new ethersV5_1.ethers.providers.JsonRpcProvider((0, context_1.rpcUrlFromNetworkConfig)(network));
        const serializedSignedTx = ethersV5_1.ethers.utils.serializeTransaction(readUnsignedEVMObject(id), (0, utils_1.prefix0x)(signature));
        const chainId = yield provider.sendTransaction(serializedSignedTx);
        return chainId.hash;
    });
}
exports.submitForDefiTxn = submitForDefiTxn;
function createUnsignedJsonObject(txHash) {
    const signatureRequest = {
        message: txHash.slice(2),
        signer: "",
    };
    const unsignedTxJson = {
        transactionType: contracts_1.contractTransactionName,
        serialization: "",
        signatureRequests: [signatureRequest],
        unsignedTransactionBuffer: "",
        usedFee: "",
        txDetails: "",
    };
    return unsignedTxJson;
}
function getContractAddress(network, contractName) {
    var _a;
    return __awaiter(this, void 0, void 0, function* () {
        const rpcUrl = (0, context_1.rpcUrlFromNetworkConfig)(network);
        const provider = new ethersV5_1.ethers.providers.JsonRpcProvider(rpcUrl);
        const abi = (0, contracts_1.getFlareContractRegistryABI)();
        if (network != "flare" && network != "costwo")
            throw new Error("Invalid network passed");
        const contract = new ethersV5_1.ethers.Contract(contracts_1.defaultContractAddresses.FlareContractRegistry[network], abi, provider);
        const result = yield contract.getContractAddressByName(contractName);
        if (result !== '0x0000000000000000000000000000000000000000')
            return result;
        const defaultAddress = (_a = contracts_1.defaultContractAddresses[contractName]) === null || _a === void 0 ? void 0 : _a[network];
        if (defaultAddress)
            return defaultAddress;
        throw new Error("Contract Address not found");
    });
}
function readUnsignedEVMObject(id) {
    const fname = `${forDefi_1.forDefiDirectory}/${forDefi_1.forDefiUnsignedTxnDirectory}/${id}.unsignedEVMObject.json`;
    if (!fs_1.default.existsSync(fname)) {
        throw new Error(`unsigned EVM Object file ${fname} does not exist`);
    }
    const serialization = fs_1.default.readFileSync(fname).toString();
    let file = JSON.parse(serialization);
    file.gasPrice = ethersV5_1.BigNumber.from(file.gasPrice.hex);
    file.gasLimit = ethersV5_1.BigNumber.from(file.gasLimit.hex);
    return file;
}
function saveUnsignedEVMObject(unsignedTx, id) {
    const fname = `${forDefi_1.forDefiDirectory}/${forDefi_1.forDefiUnsignedTxnDirectory}/${id}.unsignedEVMObject.json`;
    if (fs_1.default.existsSync(fname)) {
        throw new Error(`unsignedTx file ${fname} already exists`);
    }
    const serialization = JSON.stringify(unsignedTx, null, 2);
    fs_1.default.mkdirSync(`${forDefi_1.forDefiDirectory}/${forDefi_1.forDefiUnsignedTxnDirectory}`, { recursive: true });
    fs_1.default.writeFileSync(fname, serialization);
}
function getUnclaimedRewards(ethAddress, network) {
    return __awaiter(this, void 0, void 0, function* () {
        const rpcUrl = (0, context_1.rpcUrlFromNetworkConfig)(network);
        const validatorRewardManagerContractAddress = yield getContractAddress(network, contracts_1.validatorRewardManagerContractName);
        const provider = new ethersV5_1.ethers.providers.JsonRpcProvider(rpcUrl);
        const abi = (0, contracts_1.getValidatorRewardManagerABI)();
        const contract = new ethersV5_1.ethers.Contract(validatorRewardManagerContractAddress, abi, provider);
        const rewards = yield contract.getStateOfRewards(ethAddress);
        const totalRewardNumber = rewards[0];
        const claimedRewardNumber = rewards[1];
        const unclaimedRewards = totalRewardNumber.sub(claimedRewardNumber);
        return unclaimedRewards;
    });
}
function isOptedOut(ethAddress, network) {
    return __awaiter(this, void 0, void 0, function* () {
        const rpcUrl = (0, context_1.rpcUrlFromNetworkConfig)(network);
        const distributionToDelegatorsContractAddress = yield getContractAddress(network, contracts_1.distributionToDelegators);
        const provider = new ethersV5_1.ethers.providers.JsonRpcProvider(rpcUrl);
        const abi = (0, contracts_1.getDistributionToDelegatorsABI)();
        const contract = new ethersV5_1.ethers.Contract(distributionToDelegatorsContractAddress, abi, provider);
        const isOptedOut = yield contract.optOutCandidate(ethAddress);
        return isOptedOut;
    });
}
function signContractTransaction(wallet, unsignedTx, contract, derivationPath, transactionId, pvtKey) {
    return __awaiter(this, void 0, void 0, function* () {
        const serializedUnsignedTx = ethersV5_1.ethers.utils.serializeTransaction(unsignedTx);
        const txHash = ethersV5_1.ethers.utils.keccak256(serializedUnsignedTx);
        const unsignedTxObj = createUnsignedJsonObject(txHash);
        if (wallet === Object.keys(screen_1.walletConstants)[0]) {
            if (!derivationPath)
                throw new Error("No derivation path passed");
            const sign = yield (0, sign_1.ledgerSign)(unsignedTxObj, derivationPath);
            const serializedSignedTx = ethersV5_1.ethers.utils.serializeTransaction(unsignedTx, (0, utils_1.prefix0x)(sign.signature));
            console.log("Submitting txn to the chain");
            yield contract.provider.sendTransaction(serializedSignedTx);
        }
        else if (wallet === Object.keys(screen_1.walletConstants)[1]) {
            if (!transactionId)
                throw new Error("No transaction Id passed");
            (0, utils_1.saveUnsignedTxJson)(unsignedTxObj, transactionId);
            saveUnsignedEVMObject(unsignedTx, transactionId);
        }
        else if (wallet === Object.keys(screen_1.walletConstants)[2]) {
            if (!pvtKey)
                throw new Error("No private key passed");
            const wallet = new ethersV5_1.ethers.Wallet(pvtKey);
            const signedTx = yield wallet.signTransaction(unsignedTx);
            yield contract.provider.sendTransaction(signedTx);
        }
    });
}
////////// MIRROR FUND /////////
// fetches current validator info
const fetchValidatorInfo = (ctx) => __awaiter(void 0, void 0, void 0, function* () {
    const validator = yield ctx.pchain.getCurrentValidators();
    return validator;
});
// fetches pending validator info
const fetchPendingValidatorInfo = (ctx) => __awaiter(void 0, void 0, void 0, function* () {
    const pendingValidator = yield ctx.pchain.getPendingValidators();
    return pendingValidator;
});
// fetches the delegation stake (from both current and pending validator) for the current user
const fetchDelegateStake = (ctx, validatorFunction) => __awaiter(void 0, void 0, void 0, function* () {
    var _a, _b, _c, _d, _e, _f, _g;
    const validatorsInfo = yield validatorFunction(ctx);
    const validatorData = validatorsInfo === null || validatorsInfo === void 0 ? void 0 : validatorsInfo.validators;
    let userStake = [];
    for (let i = 0; i < validatorData.length; i++) {
        for (let j = 0; j < (validatorData[i].delegators && ((_a = validatorData[i].delegators) === null || _a === void 0 ? void 0 : _a.length)); j++) {
            if (validatorData[i].delegators[j] &&
                validatorData[i].delegators[j].rewardOwner.addresses.includes(ctx.pAddressBech32)) {
                const startDate = new Date(parseInt((_c = (_b = validatorData[i]) === null || _b === void 0 ? void 0 : _b.delegators[j]) === null || _c === void 0 ? void 0 : _c.startTime) * 1000);
                const endDate = new Date(parseInt((_e = (_d = validatorData[i]) === null || _d === void 0 ? void 0 : _d.delegators[j]) === null || _e === void 0 ? void 0 : _e.endTime) * 1000);
                userStake.push({
                    stakeAmount: parseFloat((_g = (_f = validatorData[i]) === null || _f === void 0 ? void 0 : _f.delegators[j]) === null || _g === void 0 ? void 0 : _g.stakeAmount) / 1e9,
                    startTime: startDate,
                    endTime: endDate
                });
            }
        }
    }
    return userStake;
});
// calculates the total amount of delegation
const getTotalFromDelegation = (data) => {
    let total = 0;
    for (let i = 0; i < data.length; i++) {
        total += data[i].stakeAmount;
    }
    return total;
};
/**
 * @description returns the mirror fund details
 * @param ctx - context
 * @returns - total mirror funds and funds with start and end time
 */
function fetchMirrorFunds(ctx) {
    return __awaiter(this, void 0, void 0, function* () {
        console.log("Checking your Mirror Funds ...");
        // fetch from the contract
        const rpcUrl = (0, context_1.rpcUrlFromNetworkConfig)(ctx.config.hrp);
        const pChainStakeMirrorContractAddress = yield getContractAddress(ctx.config.hrp, contracts_1.pChainStakeMirror);
        const provider = new ethersV5_1.ethers.providers.JsonRpcProvider(rpcUrl);
        const abi = (0, contracts_1.getPChainStakeMirrorABI)();
        const contract = new ethersV5_1.ethers.Contract(pChainStakeMirrorContractAddress, abi, provider);
        const stakedAmount = yield contract.balanceOf(ctx.cAddressHex);
        const stakedAmountInFLR = parseFloat((0, utils_1.integerToDecimal)(stakedAmount.toString(), 18));
        // fetch for the chain
        const delegationToCurrentValidator = yield fetchDelegateStake(ctx, fetchValidatorInfo);
        const delegationToPendingValidator = yield fetchDelegateStake(ctx, fetchPendingValidatorInfo);
        const totalDelegatedAmount = getTotalFromDelegation(delegationToCurrentValidator) +
            getTotalFromDelegation(delegationToPendingValidator);
        const totalInFLR = parseFloat(totalDelegatedAmount.toString());
        return {
            "Total Mirrored Amount": `${totalInFLR} FLR`,
            "Mirror Funds Details": Object.assign(Object.assign({}, delegationToCurrentValidator), delegationToPendingValidator)
        };
    });
}
exports.fetchMirrorFunds = fetchMirrorFunds;
/**
 * @description - function to opt you out of airdrop
 * @param {optOutOfAirdrop} params - optout parameters
 */
function optOutOfAirdrop(params) {
    return __awaiter(this, void 0, void 0, function* () {
        console.log('Opting you out of airdrop');
        const { cAddress, network, wallet, derivationPath, pvtKey, transactionId } = params;
        if (yield isOptedOut(cAddress, network)) {
            throw new Error('You have already opted out of Airdrop');
        }
        const rpcUrl = (0, context_1.rpcUrlFromNetworkConfig)(network);
        const distributionToDelegatorsContractAddress = yield getContractAddress(network, contracts_1.distributionToDelegators);
        const provider = new ethersV5_1.ethers.providers.JsonRpcProvider(rpcUrl);
        const abi = (0, contracts_1.getDistributionToDelegatorsABI)();
        const contract = new ethersV5_1.ethers.Contract(distributionToDelegatorsContractAddress, abi, provider);
        const checksumAddress = ethersV5_1.ethers.utils.getAddress(cAddress);
        const nonce = yield provider.getTransactionCount(checksumAddress);
        const config = (0, context_1.getNetworkConfig)(network);
        let gasEstimate;
        try {
            gasEstimate = yield contract.estimateGas.optOutOfAirdrop({ from: cAddress });
        }
        catch (_a) {
            console.log(chalk_1.default.red('Unable to estimate trx gas fees, Trx is likely to fail'));
            (0, process_1.exit)();
        }
        const gasPrice = yield provider.getGasPrice();
        const populatedTx = yield contract.populateTransaction.optOutOfAirdrop();
        const unsignedTx = Object.assign(Object.assign({}, populatedTx), { nonce, chainId: config.networkID, gasPrice, gasLimit: gasEstimate });
        yield signContractTransaction(wallet, unsignedTx, contract, derivationPath, transactionId, pvtKey);
    });
}
exports.optOutOfAirdrop = optOutOfAirdrop;
