"use strict";
var __awaiter = (this && this.__awaiter) || function (thisArg, _arguments, P, generator) {
    function adopt(value) { return value instanceof P ? value : new P(function (resolve) { resolve(value); }); }
    return new (P || (P = Promise))(function (resolve, reject) {
        function fulfilled(value) { try { step(generator.next(value)); } catch (e) { reject(e); } }
        function rejected(value) { try { step(generator["throw"](value)); } catch (e) { reject(e); } }
        function step(result) { result.done ? resolve(result.value) : adopt(result.value).then(fulfilled, rejected); }
        step((generator = generator.apply(thisArg, _arguments || [])).next());
    });
};
Object.defineProperty(exports, "__esModule", { value: true });
exports.getExportPCParams = exports.getImportCPParams = exports.issueSignedPvmTx = exports.getUnsignedExportTxPC = exports.getUnsignedImportTxCP = exports.exportTxPC = exports.importTxCP = void 0;
const dist_1 = require("@flarenetwork/flarejs/dist");
const platformvm_1 = require("@flarenetwork/flarejs/dist/apis/platformvm");
const utils_1 = require("@flarenetwork/flarejs/dist//utils");
const utils_2 = require("../utils");
/**
 * @description Import funds exported from C-chain to P-chain to P-chain
 * @param ctx - context with constants initialized from user keys
 */
function importTxCP(ctx, threshold) {
    return __awaiter(this, void 0, void 0, function* () {
        const params = yield getImportCPParams(ctx, threshold);
        const unsignedTx = yield ctx.pchain.buildImportTx(...params);
        const tx = unsignedTx.sign(ctx.pKeychain);
        const txid = yield ctx.pchain.issueTx(tx);
        return { txid: txid, usedFee: ctx.pchain.getDefaultTxFee().toString() };
    });
}
exports.importTxCP = importTxCP;
/**
 * @description Export funds from P-chain to C-chain.
 * @param ctx - context with constants initialized from user keys
 * @param amount - amount to export (if left undefined, it exports all funds on P-chain)
 */
function exportTxPC(ctx, amount, threshold) {
    return __awaiter(this, void 0, void 0, function* () {
        const params = yield getExportPCParams(ctx, amount, threshold);
        const unsignedTx = yield ctx.pchain.buildExportTx(...params);
        const tx = unsignedTx.sign(ctx.pKeychain);
        const txid = yield ctx.pchain.issueTx(tx);
        return { txid: txid, usedFee: ctx.pchain.getDefaultTxFee().toString() };
    });
}
exports.exportTxPC = exportTxPC;
/**
 * @description Get unsigned transaction for import from C-chain to P-chain.
 * @param ctx - context with constants initialized from user keys
 */
function getUnsignedImportTxCP(ctx, threshold) {
    return __awaiter(this, void 0, void 0, function* () {
        const params = yield getImportCPParams(ctx, threshold);
        const unsignedTx = yield ctx.pchain.buildImportTx(...params);
        return {
            transactionType: 'importCP',
            serialization: (0, utils_2.serializeUnsignedTx)(unsignedTx),
            signatureRequests: unsignedTx.prepareUnsignedHashes(ctx.cKeychain),
            unsignedTransactionBuffer: unsignedTx.toBuffer().toString('hex'),
            usedFee: ctx.pchain.getDefaultTxFee().toString()
        };
    });
}
exports.getUnsignedImportTxCP = getUnsignedImportTxCP;
/**
 * @description Get unsigned transaction for export from C-chain to P-chain.
 * @param ctx - context with constants initialized from user keys
 * @param amount - amount to export (if left undefined, it exports all funds on P-chain)
 */
function getUnsignedExportTxPC(ctx, amount, threshold) {
    return __awaiter(this, void 0, void 0, function* () {
        const params = yield getExportPCParams(ctx, amount, threshold);
        const unsignedTx = yield ctx.pchain.buildExportTx(...params);
        return {
            transactionType: 'exportPC',
            serialization: (0, utils_2.serializeUnsignedTx)(unsignedTx),
            signatureRequests: unsignedTx.prepareUnsignedHashes(ctx.cKeychain),
            unsignedTransactionBuffer: unsignedTx.toBuffer().toString('hex'),
            usedFee: ctx.pchain.getDefaultTxFee().toString()
        };
    });
}
exports.getUnsignedExportTxPC = getUnsignedExportTxPC;
/**
 * @description Issues a signed P-chain transation
 * @param ctx - context with constants initialized from user keys
 * @param id - id associated with the transaction
 */
function issueSignedPvmTx(ctx, signedTxJson) {
    return __awaiter(this, void 0, void 0, function* () {
        const signatures = Array(signedTxJson.signatureRequests.length).fill(signedTxJson.signature);
        const ecdsaSignatures = signatures.map((signature) => (0, utils_2.expandSignature)(signature));
        const unsignedTx = (0, utils_2.deserializeUnsignedTx)(platformvm_1.UnsignedTx, signedTxJson.serialization);
        const tx = unsignedTx.signWithRawSignatures(ecdsaSignatures, ctx.cKeychain);
        const chainTxId = yield ctx.pchain.issueTx(tx);
        return { chainTxId: chainTxId };
    });
}
exports.issueSignedPvmTx = issueSignedPvmTx;
/**
 * @description Returns the parameters related to importing funds from C to P chain
 * @param ctx - context file aka ctx.json
 * @param threshold
 * @returns the parameters
 */
function getImportCPParams(ctx, threshold = 1) {
    return __awaiter(this, void 0, void 0, function* () {
        const locktime = new dist_1.BN(0);
        const asOf = (0, utils_1.UnixNow)();
        const platformVMUTXOResponse = yield ctx.pchain.getUTXOs([ctx.pAddressBech32], ctx.cChainBlockchainID);
        const utxoSet = platformVMUTXOResponse.utxos;
        return [
            utxoSet,
            [ctx.pAddressBech32],
            ctx.cChainBlockchainID,
            [ctx.pAddressBech32],
            [ctx.pAddressBech32],
            [ctx.pAddressBech32],
            undefined,
            asOf,
            locktime,
            threshold
        ];
    });
}
exports.getImportCPParams = getImportCPParams;
/**
 * @description - Returns the parameters that will be used by function to export funds from P to C chain
 * @param ctx - the context file
 * @param amount - the amount that needs to be exported from P to C
 * @param threshold - threshold
 * @returns
 */
function getExportPCParams(ctx, amount, threshold = 1) {
    return __awaiter(this, void 0, void 0, function* () {
        const locktime = new dist_1.BN(0);
        const asOf = (0, utils_1.UnixNow)();
        const platformVMUTXOResponse = yield ctx.pchain.getUTXOs([ctx.pAddressBech32]);
        const utxoSet = platformVMUTXOResponse.utxos;
        const fee = ctx.pchain.getDefaultTxFee();
        // if amount is not passed, export all funds minus the fee
        if (amount === undefined) {
            const getBalanceResponse = yield ctx.pchain.getBalance(ctx.pAddressBech32);
            const unlocked = new dist_1.BN(getBalanceResponse.unlocked);
            amount = unlocked.sub(fee);
        }
        return [
            utxoSet,
            amount,
            ctx.cChainBlockchainID,
            [ctx.cAddressBech32],
            [ctx.pAddressBech32],
            [ctx.pAddressBech32],
            undefined,
            asOf,
            locktime,
            threshold
        ];
    });
}
exports.getExportPCParams = getExportPCParams;
